import h5py
import numpy as np
import matplotlib.pyplot as plt
import seaborn
import datetime
import glob


def save_dict_to_hdf5(dic, filename):
    """
    Save dictionary to h5 file

    :param dic: Python dictionary to be saved
    :param filename: Filename including path

    Dictionary may contain other dictionary, lists, scalars, strings,
    datatime and ndarray of scalars.

    Dictionaries and lists are iterated to be stored in h5 file.
    datatime is converted to integer.
    """
    with h5py.File(filename, 'w') as h5file:
        recursively_save_dict_contents_to_group(h5file, '/', dic)


def load_data_from_hdf5(filelist):
    """
    Load Saved calibration constant from h5 file

    :param filelist: List of file to be parsed with glob
    :return: dictionary with calibration constant
    """
    data = {}
    files = glob.glob(filelist)
    # Loop over files
    for filename in files:
        with h5py.File(filename, 'r') as f:
            # Loop over calibration constants
            for cKey in f.keys():
                if cKey not in data:
                    data[cKey] = {}
                # Loop over modules
                for mKey in f.get(cKey):
                    if mKey not in data[cKey]:
                        data[cKey][mKey] = {}
                        data[cKey][mKey]["mdata"] = []
                    # Loop over time stamps
                    for tKey in f.get(cKey + '/' + mKey):
                        # Loop over metadata
                        mdata = {}
                        for mdKey in f.get(
                                "/".join((cKey, mKey, tKey, 'mdata'))):
                            mdata[mdKey] = f.get(
                                "/".join(
                                    (cKey, mKey, tKey, 'mdata', mdKey))).value
                        data[cKey][mKey]["mdata"].append(mdata)

                        # Loop over stored data
                        for dKay in f.get( "/".join((cKey, mKey, tKey))):
                            if dKay == "mdata":
                                continue

                            if dKay not in data[cKey][mKey]:
                                data[cKey][mKey][dKay] = []

                            value = f.get(
                            "/".join((cKey, mKey, tKey, dKay))).value

                            if dKay == "ctime":
                                value = datetime.datetime.fromtimestamp(value)

                            data[cKey][mKey][dKay].append(value)
    return data


def recursively_save_dict_contents_to_group(h5file, path, dic):
    """
    Save dictionary to h5 file recursively

    :param h5file: h5 file
    :param path: Path in h5 file
    :param dic: Dictionary to be saved

    """
    for key, item in dic.items():
        newpath = '/'.join((path, str(key)))

        # Prepare datatime to save
        if isinstance(item, datetime.datetime):
            item = item.timestamp()

        # Iterate over dictionary
        if isinstance(item, dict):
            recursively_save_dict_contents_to_group(h5file, newpath, item)

        # Iterate over list
        elif isinstance(item, (list, tuple)):
            for i, x in enumerate(item):
                newpath = '/'.join((path, str(key), str(i)))
                recursively_save_dict_contents_to_group(h5file, newpath, x)

        # Save strings, scalars or array of scalars
        elif (np.isscalar(item) or
              isinstance(item, (str, np.ndarray))):
            h5file[newpath] = item

        # Other types cannot be saved and will result in an error
        else:
            raise ValueError('Cannot save %s type.' % type(item))


def combine_constants(ctimes_a, ctimes_b):
    """
    Combine two numpy arrays of timestamp

    :param ctimes_a: First numpy array of timestamps
    :param ctimes_b: Second numpy array of timestamps
    :return: Combined list of timestamps, Indexes of first and second lists

    Two lists of timestamps comes from two lists of calibration constants,
    which should be grouped in one. E.g. SlopesFF and SlopesPC.
    Input lists may have different number of items with may have
    intersects values. Input lists are not ordered.

    Output is a list of ordered values, which are taken from
    input lists and assigned to elements of the list of pairs.

    Another output is a list of pairs of indexes, which corresponds
    to elements of input lists, that corresponding values closely match
    each other.

    Example:
    .. code-block:: python

        # Calibration constants SlopesFF and SlopesPC retrieved
        # for a given module withoin certain time period
        cdataPC_vs_time = np.array(dataPC["data"])
        cdataFF_vs_time = np.array(dataFF["data"])

        # Timespamps from SlopesFF
        # [datetime.datetime(2018, 8, 2, 16, 32, 15)
        #  datetime.datetime(2018, 6, 6, 14, 57, 29)
        #  datetime.datetime(2018, 6, 6, 13, 6, 54)
        #  datetime.datetime(2018, 6, 6, 10, 58, 6)
        #  datetime.datetime(2018, 9, 6, 13, 24, 46)
        #  datetime.datetime(2018, 8, 23, 16, 36, 46)]
        ctimesFF = np.array(dataFF["ctime"])

        # Timespamps from SlopesPC
        # [datetime.datetime(2018, 6, 6, 13, 6, 54)
        #  datetime.datetime(2018, 6, 6, 10, 58, 52)
        #  datetime.datetime(2018, 6, 6, 9, 52, 32)
        #  datetime.datetime(2018, 9, 14, 10, 41, 9)
        #  datetime.datetime(2018, 8, 1, 10, 55, 45)]
        ctimesPC = np.array(dataPC["ctime"])

        # Get combined time stamps and indexes, which shows
        # how to combine constants
        ctime, icomb = combine_constants(ctimesFF, ctimesPC)

        # ctime values:
        # [datetime.datetime(2018, 6, 6, 9, 52, 32),
        #  datetime.datetime(2018, 6, 6, 10, 58, 6),
        #  datetime.datetime(2018, 6, 6, 10, 58, 52),
        #  datetime.datetime(2018, 6, 6, 13, 6, 54),
        #  datetime.datetime(2018, 6, 6, 13, 6, 54),
        #  datetime.datetime(2018, 6, 6, 14, 57, 29),
        #  datetime.datetime(2018, 8, 1, 10, 55, 45),
        #  datetime.datetime(2018, 8, 2, 16, 32, 15),
        #  datetime.datetime(2018, 8, 23, 16, 36, 46),
        #  datetime.datetime(2018, 9, 6, 13, 24, 46)]

        # icomb:
        # [(3, 2), (3, 1), (2, 1), (2, 4), (2, 4), (1, 4), (0, 4),
        #  (0, 3), (5, 3), (4, 3)]

        # create a list of gain constants by combining SlopesPC and FF
        gain_vs_time = []
        for iFF, iPC in icomb:
            gain_vs_time.append(cdataFF_vs_time[iFF] * cdataPC_vs_time[iPC])

    """
    sort_ind_a = np.argsort(ctimes_a)
    sort_ind_b = np.argsort(ctimes_b)

    t_asort = ctimes_a[sort_ind_a]
    t_bsort = ctimes_b[sort_ind_b]

    t_sort = np.array(sorted(np.hstack((t_asort, t_bsort))))

    icomb = []
    ctime = []
    for time in t_sort:
        if len(ctime) == len(t_sort) - 1:
            break
        ctime.append(time)
        if time in t_asort:
            i_a = np.where(t_asort == time)[0][0]
            if time >= t_bsort[-1]:
                i_b = len(t_bsort) - 1
            else:
                i_b = np.where(t_bsort > time)[0][0]
        else:
            if time >= t_asort[-1]:
                i_a = len(t_asort) - 1
            else:
                i_a = np.where(t_asort > time)[0][0]
            i_b = np.where(t_bsort == time)[0][0]
        icomb.append((sort_ind_a[i_a], sort_ind_b[i_b]))

    return ctime, icomb


def plot_heatmap(data, x_title='Pixel X', y_title='Pixel Y', x_labels='',
                 title='Average over pixels and cells', label='', fname='',
                 vmax=None, vmin=None, ytickspos=None, yticks=None):
    """Plotting heatmap from 2D data

    Args:
        data: 2D dataset.
        x_title: Label for x axis
        y_title: Label for y axis
        title: Title on the top side of a plot
        x_labels: Ticks for X axis (list or array)
        yticks: Ticks for Y axis (list or array)
        ytickspos: Positions of ticks for Y axis (list or array)
        vmax: Maximum value of heatmap
        vmin: Minimum value of heatmap
        label: Label on the right side of a plot
        fname: Name of output file (including path)
    """

    fig = plt.figure(figsize=(12, 12))
    ax = fig.add_subplot(111)

    if vmin is None:
        vmin = np.nanmin(data[np.nonzero(data)])

    seaborn.heatmap(data, xticklabels=10, yticklabels=32, cmap='jet',
                    vmin=vmin, vmax=vmax, cbar_kws={"label": label})
    ax.set_xticklabels(x_labels, minor=False, rotation=90)
    ax.figure.axes[-1].yaxis.label.set_size(20)

    plt.xticks(np.arange(data[0, :].size) + 0.3, )

    plt.xlabel(x_title, size=18)
    plt.ylabel(y_title, size=18)
    plt.title(title, size=18)

    if ytickspos is not None:
        plt.yticks(ytickspos, yticks)

    plt.subplots_adjust(wspace=0.2, hspace=0.2)

    if fname != '':
        fig.savefig(fname, bbox_inches='tight')


# The following functions are used to combine two lists in one.
# Elements present in both lists counted only once.
# Output list is sorted.

def to_multiset(x):
    """
    Convert list to multiset

    :param x: List to be converted
    :return: ordered multiset: Item of list, index of this element

    E.g. (2,3,3,4,2) gives {(2, 0), (2, 1), (3, 0), (3, 1), (4, 0)}
    """
    result = set()
    max_rep = len(x)
    for elt in x:
        for n in range(max_rep):
            n_elt = (elt, n)
            if n_elt not in result:
                result.add(n_elt)
                break
    return result


def from_multiset(x):
    """
    Convert multiset to sorted list

    :param x: Multi-set to be converted
    :return: Sorted list
    """
    return sorted([elt for elt, n in x])


def multi_union(a, b):
    """
    Combine two lists in one.
    If the same element is present in both lists, it counts only once.

    :param a: First list
    :param b: Second list
    :return: Combined list.

        Example:
    .. code-block:: python

        a = [5,5]
        b = [4,4,4,5]
        comb = multi_union(a, b)
        # comb: [4,4,4,5,5]

    """
    aa = to_multiset(a)
    bb = to_multiset(b)
    return from_multiset(aa | bb)


def multi_intersect(a, b):
    """
    Combine two lists in one. Keep only elements present in both lists.

    :param a: First list
    :param b: Second list
    :return: Combined list.
    """
    aa = to_multiset(a)
    bb = to_multiset(b)
    return from_multiset(aa & bb)
