import argparse
from collections import OrderedDict
import copy
import glob
from jinja2 import Template
import os
from os.path import isfile, isdir, splitext
from queue import Queue
import re
import shutil
from subprocess import Popen, PIPE, check_output, check_call
import textwrap
from time import sleep
import time
from uuid import uuid4


def atoi(text):
    '''
    Convert string to integer is possible

    :param text: string to be converted
    :return: integer value or input string
    '''

    return int(text) if text.isdigit() else text


def natural_keys(text):
    '''
    Decompose string to list of integers and sub-strings
    '''
    return [ atoi(c) for c in re.split(r'(\d+)', text) ]


def combine_report(run_path, calibration):
    sphinx_path = "{}/sphinx_rep".format(os.path.abspath(run_path))
    os.makedirs(sphinx_path)
    direntries = os.listdir(run_path)
    direntries.sort(key=natural_keys)
    
    for entry in direntries:
        
        if isfile("{}/{}".format(run_path, entry)):
            name, ext = splitext("{}".format(entry))
            
            if ext == ".rst":
                comps = name.split("__")                
                if len(comps) >= 3:
                    group, name_param, conc_param = "_".join(comps[:-2]), comps[-2], comps[-1]
                else:
                    group, name_param,  conc_param = comps[0], "None", "None"

                with open("{}/{}.rst".format(sphinx_path, group), "a") as gfile:
                    if conc_param != "None":
                        title = "{}. {} = {}".format(calibration, name_param, conc_param)
                        gfile.write(title + "\n")
                        gfile.write( "=" *len (title) + "\n")
                        gfile.write("\n")
                    with open("{}/{}".format(run_path, entry), "r") as ifile:
                        skip_next = False
                        for line in ifile.readlines():
                            if skip_next:
                                skip_next = False
                                continue
                            if conc_param == "None":
                                gfile.write(line)
                            elif " ".join(calibration.split()) != " ".join(line.split()):
                                gfile.write(line)
                            else:
                                skip_next = True
                                
                    gfile.write("\n\n")
        if isdir("{}/{}".format(run_path, entry)):
            shutil.copytree("{}/{}".format(run_path, entry), "{}/{}".format(sphinx_path, entry))
    return sphinx_path


def prepare_plots(run_path, threshold=1000000):
    """
    Convert svg file to pdf or png to be used for latex

    Conversion of svg to vector graphics pdf is performed using svglib3.
    This procedure is CPU consuming. In order to speed up process
    large svg files are converted to png format.

    The links in the rst files are adapted accordingly to the
    converted image files.

    :param run_path: Run path of the slurm job
    :param threshold: Max svg file size (in bytes) to be converted to pdf
    """
    print('Convert svg to pdf and png')
    run_path = os.path.abspath(run_path)

    rst_files = glob.glob('{}/*rst'.format(run_path))
    for rst_file in rst_files:
        rst_file_name = os.path.basename(rst_file)
        rst_file_name = os.path.splitext(rst_file_name)[0]

        svg_files = glob.glob(
            '{}/{}_files/*svg'.format(run_path, rst_file_name))
        for f_path in svg_files:
            f_name = os.path.basename(f_path)
            f_name = os.path.splitext(f_name)[0]

            if (os.stat(f_path)).st_size < threshold:
                check_call(["svg2pdf", "{}".format(f_path)], shell=False)
                new_ext = 'pdf'
            else:
                check_call(["convert", "{}".format(f_path),
                            "{}.png".format(f_name)], shell=False)
                new_ext = 'png'

            check_call(["sed",
                        "-i",
                        "s/{}.svg/{}.{}/g".format(f_name, f_name, new_ext),
                        rst_file],
                       shell=False)


def make_timing_summary(run_path, joblist):
    """
    Create an rst file with timing summary of executed slurm jobs

    :param run_path: Run path of the slurm job
    :param joblist: List of slurm jobs
    """
    print('Prepare timing summary')
    run_path = os.path.abspath(run_path)
    pars_vals = []
    pars = 'JobID,Elapsed,Suspended'
    pars_name = pars.split(',')

    for job in joblist:
        out = check_output(['sacct', '-j', job,
                            '--format={}'.format(pars)],
                           shell=False)
        l = str(out).split('\\n')

        # loop over output lines, skip first two lines with header.
        for line in l[2:]:
            s = line.split()
            if len(s) == len(pars_name):
                pars_vals.append(s)
                break

    tmpl = Template('''
                    Timing summary
                    ==============
                    
                    .. math::
                        {% for line in table %}
                        {{ line }}
                        {%- endfor %}
                    ''')
    import tabulate
    with open("{}/timing_summary.rst".format(run_path), "w+") as gfile:

        table = tabulate.tabulate(pars_vals, tablefmt='latex',
                                  headers=pars_name)
        gfile.write(textwrap.dedent(tmpl.render(table=table.split('\n'))))


def make_report(run_path, tmp_path, out_path, project, author, version, report_to):
    run_path = os.path.abspath(run_path)
    report_path, report_name = os.path.split(report_to)
    if report_path != '':
        out_path = report_path

    try:
        import subprocess
        subprocess.check_call(["sphinx-quickstart",
                               "--quiet",
                               "--project='{}'".format(project),
                               "--author='{}'".format(author),
                               "-v", str(version),
                               "--suffix=.rst",
                               "--master=index",
                               "--ext-intersphinx",
                               "--ext-mathjax",
                               "--makefile",
                               "--no-batchfile", run_path])

    except subprocess.CalledProcessError:
        raise Exception("Failed to run sphinx-quickbuild. Is sphinx installed?"
                        "Generated simple index.rst instead")

    # quickbuild went well we need to edit the index.rst and conf.py files
    from shutil import move

    module_path = "{}".format(os.path.abspath(os.path.dirname(__file__)))
    with open("{}/conf.py.tmp".format(run_path), "w") as mf:
        with open("{}/conf.py".format(run_path), "r") as mfr:
            for line in mfr:
                if line == '# latex_logo = None\n':
                    mf.write("latex_logo = '{}/xfel.jpg'\n".format(module_path))
                    mf.write("latex_additional_files = ['titlepage.tex.txt']\n")

                elif line == 'latex_elements = {\n':
                    mf.write(line)
                    mf.write("    'extraclassoptions': ',openany, oneside',\n")
                    mf.write("    'maketitle': r'\input{titlepage.tex.txt}',\n")

                elif '{}.tex'.format(project.replace(" ", "")) in line:
                    line.replace('{}.tex'.format(project.replace(" ", "")),
                                 '{}.tex'.format(report_name))
                    mf.write(line)

                else:
                    mf.write(line)

    _ = os.getcwd()
    os.remove("{}/conf.py".format(run_path))
    move("{}/conf.py.tmp".format(run_path), "{}/conf.py".format(run_path))

    direntries = os.listdir(run_path)
    files_to_handle = []
    for entry in direntries:
        if isfile("{}/{}".format(run_path, entry)):
            name, ext = splitext("{}".format(entry))            
            if ext == ".rst" and "index" not in name:
                files_to_handle.append(name.strip())

    index_tmp = Template('''
                        Calibration report
                        ==================
                        
                        .. toctree::
                           :maxdepth: 2
                           {% for k in keys %}
                           {{ k }}
                           {%- endfor %}
                        ''')

    with open("{}/index.rst".format(run_path), "w+") as mf:
        mf.write(textwrap.dedent(index_tmp.render(keys=files_to_handle)))

    # finally call the make scripts
    os.chdir(run_path)
    try:
        import subprocess
        subprocess.check_call(["make", "latexpdf"])

    except subprocess.CalledProcessError:
        print("Failed to make pdf documentation")
        print("Temp files will not be deleted and "+
              "can be inspected at: {}".format(run_path))
        return
    print("Moving report to final location: {}".format(out_path))
    for f in glob.glob(r'{}/_build/latex/*.pdf'.format(run_path)):
        print(f)
        shutil.copy(f, out_path)
    print("Removing temporary files at: {}".format(tmp_path))
    shutil.rmtree(tmp_path)

    
def make_titlepage(sphinx_path, project, data_path):
    with open("{}/titlepage.tex.txt".format(sphinx_path), "w+") as gfile:
        gfile.write('\\begin{titlepage}\n')
        gfile.write('  \\centering\n')
        gfile.write('  \\vspace*{40mm}\n')
        gfile.write('  \\textbf{\\Huge {')
        gfile.write(project)
        gfile.write('}}\n')
        gfile.write('  \\vspace{0mm}\n\n')
        gfile.write('  \\begin{figure}[!h]\n')
        gfile.write('  \\centering\n')
        gfile.write('  \\includegraphics[scale=0.3]{xfel.jpg}\n')
        gfile.write('  \\end{figure}\n\n')
        gfile.write('  \\vspace{0mm}\n')
        gfile.write('  \\Large \\textbf{{Detector group}}\n\n')
        gfile.write('  Based on data sample: {}\n\n'.format(data_path))
        gfile.write('  Release : 0.1\n\n')
        gfile.write('  \\vspace*{0mm}\n')
        gfile.write('  \\today\n')
        gfile.write('  \\vfill\n')
        gfile.write('  \\textit{For internal use only}\n')
        gfile.write('\\end{titlepage}\n')
        gfile.write("\n\n")


def finalize(joblist, finaljob, run_path, out_path, project, calibration,
             author, version, report_to, data_path='Unknown'):
    
    print("Waiting on jobs to finish: {}".format(joblist))
    while True:
        found_jobs = set()
        output = check_output(['squeue']).decode('utf8')
        for line in output.split("\n"):
            for job in joblist:
                if str(job) in line:
                    found_jobs.add(job)
        if len(found_jobs) == 0:
            break
        sleep(10)

    prepare_plots(run_path)
    make_timing_summary(run_path, joblist+[str(finaljob)])
    sphinx_path = combine_report(run_path, calibration)
    make_titlepage(sphinx_path, project, data_path)
    make_report(sphinx_path, run_path, out_path, project, author, version,
                report_to)


def parse_runs(runs, return_type=str):
    pruns = []
    if isinstance(runs, str):
        for rcomp in runs.split(","):
            if "-" in rcomp:
                start, end = rcomp.split("-")
                pruns += list(range(int(start), int(end)))
            else:
                pruns += [int(rcomp)]
    elif isinstance(runs, (list, tuple)):
        pruns = runs
    else:
        pruns = [runs,]
        
    if return_type is str:
        return ["r{:04d}".format(r) for r in pruns]
    else:
        return pruns
                

def run_prop_seq_from_path(filename):
    run = re.findall(r'.*r([0-9]{4}).*', filename)
    run = run[0] if len(run) else None
    proposal = re.findall(r'.*p([0-9]{6}).*', filename)
    proposal = proposal[0] if len(proposal) else None
    sequence = re.findall(r'.*S([0-9]{5}).*', filename)
    sequence = sequence[0] if len(sequence) else None
    return run, proposal, sequence


def map_modules_from_files(filelist, file_inset, quadrants, modules_per_quad):
    total_sequences = 0
    total_file_size = 0
    module_files = {}
    mod_ids = {}
    for quadrant in range(0, quadrants):
        for module in range(0, modules_per_quad):
            name = "Q{}M{}".format(quadrant + 1, module + 1)
            module_files[name] = Queue()
            num = quadrant * 4 + module
            mod_ids[name] = num
            file_infix = "{}{:02d}".format(file_inset, num)
            for file in filelist:
                if file_infix in file:
                    module_files[name].put(file)
                    total_sequences += 1
                    total_file_size += os.path.getsize(file)
                    
    return module_files, mod_ids, total_sequences, total_file_size


def gain_map_files(in_folder, runs, sequences, file_inset, quadrants, modules_per_quad):
    total_sequences = 0
    total_file_size = 0
    gain_mapped_files = OrderedDict()
    for gain, run in runs.items():
        ginfolder = "{}/{}".format(in_folder, run)
        dirlist = os.listdir(ginfolder)
        file_list = []
        for entry in dirlist:
            #only h5 file
            abs_entry = "{}/{}".format(ginfolder, entry)
            if os.path.isfile(abs_entry) and os.path.splitext(abs_entry)[1] == ".h5":

                if sequences is None:
                    file_list.append(abs_entry)
                else:
                    for seq in sequences:
                        if "{:05d}.h5".format(seq) in abs_entry:
                            file_list.append(os.path.abspath(abs_entry))
                            
        mapped_files, mod_ids, seq, fs = map_modules_from_files(file_list, file_inset,
                                                                quadrants, modules_per_quad)
        total_sequences += seq
        total_file_size += fs
        gain_mapped_files[gain] = mapped_files
    return gain_mapped_files, total_sequences, total_file_size/1e9


# from https://github.com/jupyter/notebook/issues/1000
import json
import os.path
import re
import ipykernel
import requests

try:  # Python 3
    from urllib.parse import urljoin
except ImportError:  # Python 2
    from urlparse import urljoin

try:  # Python 3
    from notebook.notebookapp import list_running_servers
except ImportError:  # Python 2
    import warnings
    from IPython.utils.shimmodule import ShimWarning
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=ShimWarning)
        from IPython.html.notebookapp import list_running_servers


def get_notebook_name():
    """
    Return the full path of the jupyter notebook.
    """
    try:
        kernel_id = re.search('kernel-(.*).json',
                              ipykernel.connect.get_connection_file()).group(1)
        servers = list_running_servers()
        for ss in servers:
            response = requests.get(urljoin(ss['url'], 'api/sessions'),
                                    params={'token': ss.get('token', '')})
            for nn in json.loads(response.text):
                if nn['kernel']['id'] == kernel_id:
                    relative_path = nn['notebook']['path']
                    return relative_path
    except:
        return os.environ.get("CAL_NOTEBOOK_NAME", "Unknown Notebook")
    
    
def get_dir_creation_date(directory, run):
    import os
    import datetime
    creation_time = os.stat("{}/r{:04d}".format(directory, run)).st_ctime 
    creation_time = datetime.datetime.fromtimestamp(creation_time)
    return(creation_time)


def get_dir_creation_date(directory, run):
    import os
    import datetime
    creation_time = os.stat("{}/r{:04d}".format(directory, run)).st_mtime 
    creation_time = datetime.datetime.fromtimestamp(creation_time)
    return(creation_time)


already_printed = {}
def get_from_db(device, constant, condition, empty_constant,
                         cal_db_interface, creation_time = None,
                         verbosity=1, timeout=30000):
    """
    Return calibration constants and metadata requested from CalDB

    :param device: Instance of detector
    :param constant: Calibration constant known for given detector
    :param condition: Calibration condition
    :param empty_constant: Constant to be returned in case of failure.
    :param cal_db_interface: Interface string, e.g. "tcp://max-exfl016:8015"
    :param creation_time: Latest time for constant to be created
    :param verbosity: Level of verbosity (0 - silent)
    :param timeout: Timeout for zmq request
    :return: Calibration constant, metadata
    """
    from iCalibrationDB import ConstantMetaData, Constants, Conditions, Detectors, Versions
    
    if device:
        metadata = ConstantMetaData()        
        metadata.calibration_constant = constant
        metadata.detector_condition = condition
        if creation_time is None:
            metadata.calibration_constant_version = Versions.Now(device=device)
        else:
            metadata.calibration_constant_version = Versions.Timespan(device=device,
                                                                      start=creation_time)
            creation_time = creation_time.isoformat()

        try:
            metadata.retrieve(cal_db_interface, when=creation_time, timeout=timeout)
            if verbosity > 0:
                if constant.name not in already_printed or verbosity > 1:
                    already_printed[constant.name] = True
                    print("{} was injected on: {}".format(constant.name,
                                                          metadata.calibration_constant_version.begin_at))
            return constant.data, metadata
        except Exception as e:
            if verbosity > 0:
                print(e)
            return empty_constant, metadata
    else:
        return empty_constant, None


def get_constant_from_db(device, constant, condition, empty_constant,
                         cal_db_interface, creation_time=None,
                         print_once=True, timeout=30000):
    """
    Return calibration constants requested from CalDB
    """
    data, _ = get_from_db(device, constant, condition, empty_constant,
                               cal_db_interface, creation_time,
                               int(print_once), timeout)

    return data
